/* jshint undef: true, unused: true */
/* global require, module, com, org */
var _ = require("lodash");
var Objects = require("convenience").Objects;
var request = require("request");
var StringUtilities = com.cattura.cattura_commons.utilities.StringUtilities;

/*
 * Flow:
 *  * 0: (Optional) CCCP /cranium/authorize?cranium=<Cranium URL>
 *  - 1: Cranium /unit/add?redirect_uri=<redirect URI>
 *  * 2: CCCP /cranium/authorize?code=<code from Cranium>&unitId<unit id from Cranium>&cranium=<Cranium URL>
 *  - 3: Cranium /unit/<unit id>?code=<code from CCCP>
 *  - 4: Redirect URL
 */
function setup (options)
{
    var server = options.server;
    var authorizeTemplate = _(options.authorizeTemplate.read()).template();
    var craniums = options.craniums;
    var addCranium = options.addCranium;
    
    var flows = {}; // Requests to connect with Cranium that need to be confirmed.
    var accessTokens = {}; // Auth codes that we have generated for Cranium.
    
    var HttpHandler = org.webbitserver.HttpHandler;    
    server.add("/cranium/authorize", new HttpHandler({
       handleHttpRequest: function (req, res) {
           var requestId = req.postParam("requestId") || req.queryParam("requestId"), // Internal request id.
               craniumUrl = Objects.toJS(req.postParam("cranium") || req.queryParam("cranium")), // URL of Cranium.
               unitUri = Objects.toJS(req.postParam("unit_uri") || req.queryParam("unit_uri")), // Unit URI known by the UI and provided to Cranium when setting up a unit.
               craniumCode = req.postParam("code") || req.queryParam("code"), // Authorization code provided by Cranium, if we were sent here from there.
               redirectUri = req.postParam("redirect_uri") || req.queryParam("redirect_uri"), // User-specified redirect URL
               unitId = req.postParam("unitId") || req.queryParam("unitId"), // Unit ID known by Cranium
               flow;

           /*
            * If we're in the middle of an existing flow, then get the values
            * from that flow.
            */
           if (requestId in flows)
           {
               flow = flows[requestId];

               if ('craniumCode' in flow)
               {
                   craniumCode = flow.craniumCode;
               }

               if ('craniumUrl' in flow)
               {
                   craniumUrl = flow.craniumUrl;
               }
           }            

           /*
            * Sanitize the cranium url by removing the slash.
            */
           if (craniumUrl && craniumUrl.charAt(craniumUrl.length - 1) === "/")
           {
               craniumUrl = craniumUrl.substring(0, craniumUrl.length - 1);
           }

           /*
            * (2) We have a code from Cranium, and...
            */
           if (craniumCode)
           {
               /*
                * ... no flow on our side, so create a flow and prompt the user for confirmation.
                */
               if (!flow)
               {
                   requestId = StringUtilities.createRandomString(32, "abcdefghijklmnopqrstuvwxyz0123456789");
                   flow = {
                       craniumCode: craniumCode,
                       craniumUrl: craniumUrl,
                       redirectUri: redirectUri // Callback provided at the start of the flow (usually either to CCCP UI or Cranium)
                   };

                   flows[requestId] = flow;
                   res.content(authorizeTemplate({
                       requestId: requestId,
                       request: flow,
                       error: null
                   })).end();
               }
               /*
                * ... a completed flow (the user has confirmed), so go back to
                * Cranium and finish the flow there.
                */
               else
               {
                   flow.craniumCode = craniumCode;

                   /*
                    * Flow is finished on our end.
                    */
                   delete flows[requestId];

                   /*
                    * Generate an authorization code for Cranium, as well as an access token to exchange it for.
                    */
                   var code = StringUtilities.createRandomString(32, "abcdefghijklmnopqrstuvwxyz0123456789");
                   accessTokens[code] = StringUtilities.createRandomString(32, "abcdefghijklmnopqrstuvwxyz0123456789");

                   /*
                    * Request access token from Cranium using our authorization code.
                    * This also verifies that we can access Cranium from the unit end.
                    * If Cranium gives us the access token, we redirect back to Cranium
                    * to finish the flow.
                    */
                   request.post(flow.craniumUrl + "/access_token", {
                       body: {
                           code: flow.craniumCode
                       },
                       json: true
                   }, function (error, response, body) {
                       /*
                        * If we get an error, display error message to user.
                        */
                       if (error || 'err' in body)
                       {
                           var message = "An error occurred.";
                           if (error && 'javaException' in error)
                           {
                               error.javaException.printStackTrace();
                               message = error.javaException.message;
                           }
                           else if ('err' in body)
                           {
                               message = body.err;
                           }

                           res.status(400).content(authorizeTemplate({
                               requestId: requestId,
                               error: message,
                               request: flow
                           })).end();
                           return;
                       }

                       var cranium = addCranium({
                           url: flow.craniumUrl, 
                           outboundAccessToken: body.token, // Token used by unit to talk to Cranium.
                           inboundAccessToken: accessTokens[code] // Token used by Cranium to talk to unit.
                       });

                       /*
                        * Push initial status updates to Cranium
                        */
                       cranium.pushInitialUpdate().then(function () {
                           /*
                            * Set initial value of available to true
                            * since we successfully pushed an update
                            */
                           cranium.setAvailable(true);

                           /*
                            * Redirect back to Cranium to finish the flow.
                            */
                           res.status(307).header("Location", flow.craniumUrl + "/units/" + unitId + "?code=" + encodeURIComponent(code) + "&redirect_uri=" + encodeURIComponent(flow.redirectUri)).end();
                       });
                   });
               }
           }
           /*
            * (0) We have no Cranium code, but we do have a flow, so redirect to Cranium to 
            * get the code, then Cranium will redirect back to us.
            */
           else if (flow)
           {
               /*
                * If we are already connected to this Cranium, it is a bad request. Kill the flow
                * and render an error message.
                */
               if (_(craniums).find(function (cranium) { return cranium.url == craniumUrl; }))
               {
                   delete flows[requestId];
                   res.status(400).content(authorizeTemplate({
                       requestId: requestId,
                       error: "Already connected to Cranium instance at " + craniumUrl,
                       request: flow
                   })).end();
                   return;
               }

               flow.craniumUrl = craniumUrl;
               flow.redirectUri = redirectUri;

               res.status(307).header("Location", craniumUrl + "/units?unit_uri=" + encodeURIComponent(unitUri) + "&redirect_uri=" + encodeURIComponent(redirectUri)).end();
           }
           /*
            * Starting the flow on the CCCP side. Generate a flow and ask the user
            * to enter the Cranium URL.
            */
           else
           {
               requestId = StringUtilities.createRandomString(32, "abcdefghijklmnopqrstuvwxyz0123456789");
               flows[requestId] = {};
               res.content(authorizeTemplate({
                   requestId: requestId,
                   request: flows[requestId],
                   error: null
               })).end();
           }
       }
   }));

   /*
    * Endpoint that the Cranium server uses to exchange an authorization
    * code for an access token.
    */
   server.add("/cranium/access_token", new HttpHandler({
       handleHttpRequest: function (req, res) {
           var code = req.queryParam('code') || req.postParam('code'),
               accessToken = accessTokens[code];

           if (!code || !accessToken)
           {
               res.status(400).header("Content-Type", "application/json").content(JSON.stringify({err: "Invalid code"})).end();
               return;
           }

           delete accessTokens[code];
           res.status(200).header("Content-Type", "application/json").content(JSON.stringify(Objects.toJS({token: accessToken}, true))).end();
       }
   }));
}

module.exports = setup;