/* global -Promise */
var _ = require("lodash");
var request = require("request");
var status = require("lib/status");
var schedule = require("lib/schedule");
var Promise = require("rsvp").Promise;
var console = require("console")();
var Objects = require("convenience").Objects;

var cJSON = com.cattura.share.utils.json.parsing.JSON.JSON;
var AtomicBoolean = java.util.concurrent.atomic.AtomicBoolean;
var ExceptionUtils = com.cattura.cattura_commons.utilities.exception.ExceptionUtils;
var User = com.cattura.agents.configuration_module.model.interfaces.User;

var http = new (require("lib/http-alt"))();
    
/*
 * The features object consists of features this version of
 * the Cranium Integration Plugin supports. Cranium can inspect this
 * to determine what features are supported and what versions of features
 * are supported.
 * 
 * For versions of Cranium Integration Plugin >= 1.1.0, this should be assumed
 * to be {authv1: 1, internalApi: 1}.
 */
var FEATURES = {
    /*
     * Auth flow v1. This is the legacy auth flow. Auth flow v2 is preferred.
     */
    authv1: 1,
    
    /*
     * Auth flow v2. This is the "current" auth flow.
     */
    authv2: 1,
    
    /*
     * The internal API that Cranium and CaptureCast use to talk. This is versioned
     * on both sides.
     */
    internalApi: 1,
    
    /*
     * Can Cranium authenticate against the external CaptureCast API?
     */
    externalApi: 1
};

function Cranium (data)
{
    var t = _(this).extend(data).value();
    var user = new User({
        getUsername: function getUsername ()
        {
            return "Cranium at " + t.url;
        },
        getRoles: function getRoles ()
        {
            return [User.Role.ADMINISTRATOR];
        },
        isRole: function isRole (roleName)
        {
            return roleName == "Administrator";
        },
        can: function can ()
        {
            return true;
        }
    });
    
    this.available = new AtomicBoolean();
    this.asUser = function asUser () { return user; };
}

function makeAPIRequest (options)
{
    var cranium = options.cranium,
        endpoint = options.endpoint || "/units/me",
        method = options.method || "get",
        data = options.data || {},
        json = 'json' in options ? options.json : true;
    
    return new Promise(function (resolve, reject) {
        request({
            uri: cranium.url + endpoint,
            method: method,
            body: data,
            headers: {
                Authorization: "bearer " + cranium.outboundAccessToken
            },
            json: json,
            client: http
        }, function (err, response, body) {
            if (cranium.isAvailable() && (err || response.statusCode >= 500))
            {
                cranium.setAvailable(false);
                console.log("Cranium at " + cranium.url + " is no longer available");
            }

            if (err)
            {
                reject(err);
            }
            else if (response.statusCode >= 400)
            {
                reject(new java.lang.Exception("Status code: " + response.statusCode));
            }
            else
            {
                if (!cranium.isAvailable())
                {
                    cranium.setAvailable(true);
                    console.log("Cranium at " + cranium.url + " is now available");
                }

                resolve(body);
            }
        });
    });
}

Cranium.prototype.healthCheck = function healthCheck ()
{
    var t = this;
    console.log("Health checking Cranium at " + this.url);
    
    return makeAPIRequest({
        cranium: this,
        endpoint: "/health",
        method: "get",
        json: false
    }).then(function (result) {
        t.setAvailable(true);
    }).catch(function (err) {
        console.log("Health check failed: " + err);
        t.setAvailable(false);
    });
};

Cranium.prototype.checkAvailability = function checkAvailability ()
{
    var t = this;
    console.log("Checking availability of Cranium at " + this.url);
    
    return makeAPIRequest({
        cranium: this,
        endpoint: "/units/me",
        method: "get",
        json: false
    }).then(function (result) {
        t.setAvailable(true);
    }).catch(function (err) {
        console.log("Availability check failed: " + err);
        t.setAvailable(false);
    });
};

Cranium.prototype.pushStatusUpdate = function pushStatusUpdate (items)
{
    console.log("Pushing status update to " + this.url);
    
    var data = {
        err: null
    };

    if (!items)
    {
        items = ["info", "network", "storage", "features"];
    }
    
    if (_(items).includes("info"))
    {
        data.info = status.getUnitInfo();
    }

    if (_(items).includes("network"))
    {
        data.network = status.getNetworkInfo();
    }

    if (_(items).includes("storage"))
    {
        data.storage = status.getDiskSpaceInfo();
    }
    
    if (_(items).includes("features"))
    {
        data.features = FEATURES;
    }    
    
    return makeAPIRequest({
        cranium: this,
        endpoint: "/units/me",
        method: "patch",
        data: data
    });
};

Cranium.prototype.pushInitialUpdate = function pushInitialUpdate ()
{
    var t = this;
    return t.pushStatusUpdate().then(function () {
        return t.pushSettingsUpdate();
    }).then(function () {
        return t.pushDevicesUpdate();
    }).then(function () {
        return t.pushSchedulerConfigUpdate();
    }).then(function () {
        return t.pushScheduleUpdate();
    }).then(function () {
        return t.pushCaptureSessionsUpdate();
    }).then(function () {
        return t.pushTemplatesUpdate();
    });
};

Cranium.prototype.pushDevicesUpdate = function pushDevicesUpdate ()
{
    console.log("Pushing devices update to " + this.url);
    return makeAPIRequest({
        cranium: this,
        endpoint: "/units/me/devices",
        method: "put",
        data: status.getConnectedDevices()
    });
};

Cranium.prototype.pushTemplatesUpdate = function pushTemplatesUpdate ()
{
    console.log("Pushing templates update to " + this.url);
    return makeAPIRequest({
        cranium: this,
        endpoint: "/units/me/templates",
        method: "put",
        data: status.getTemplates()
    });
};

Cranium.prototype.pushSettingsUpdate = function pushSettingsUpdate ()
{
    console.log("Pushing settings update to " + this.url);
    return makeAPIRequest({
        cranium: this,
        endpoint: "/units/me/settings",
        method: "put",
        data: status.getSettings()
    });
};

Cranium.prototype.pushScheduleUpdate = function pushScheduleUpdate ()
{
    console.log("Pushing schedule update to " + this.url);
    return makeAPIRequest({
        cranium: this,
        endpoint: "/units/me/schedule",
        method: "put",
        data: schedule.getSchedule()
    });
};

Cranium.prototype.pushScheduleEventUpdate = function pushScheduleUpdate (eventObject)
{
    console.log("Pushing single event update to " + this.url);
    return makeAPIRequest({
        cranium: this,
        endpoint: "/units/me/schedule/" + eventObject.id,
        method: "put",
        data: eventObject
    });
};

Cranium.prototype.pushCaptureSessionsUpdate = function pushCaptureSessionsUpdate ()
{
    console.log("Pushing capture sessions update to " + this.url);
    return makeAPIRequest({
        cranium: this,
        endpoint: "/units/me/sessions",
        method: "put",
        data: status.getCaptureSessions()
    });
};

Cranium.prototype.pushSchedulerConfigUpdate = function pushSchedulerConfigUpdate ()
{
    console.log("Pushing scheduler config update to " + this.url);
    return makeAPIRequest({
        cranium: this,
        endpoint: "/units/me/schedulerConfig",
        method: "put",
        data: schedule.getSchedulerConfig()
    });
};

Cranium.prototype.pushEvent = function pushEvent (event)
{
    console.log("Pushing " + event.type + " (" + event.getClass().getSimpleName() + ") event to " + this.url);
    
    /*
     * Some Java objects can't be serialized with Rhino's JSON implementation
     * so we use ours which is backed by Jackson
     */
    var eventData = cJSON.stringify(event.data).getVal();
    
    if (!eventData)
    {
        console.log("Could not serialize " + event.type + " event");
        return Promise.reject();
    }
    
    eventData = JSON.parse(eventData);
    
    return makeAPIRequest({
        cranium: this,
        endpoint: "/units/me/events",
        method: "post",
        data: Objects.toJS({
            type: {
                name: event.type.name(),
                label: event.type.label,
                category: event.type.category.name(),
                severity: event.type.severity.name()
            },
            description: event.description,
            time: event.time.time,
            data: eventData
        }, true)
    });
};

Cranium.prototype.pushIncident = function pushEvent (incident)
{
    console.log("Pushing incident event to " + this.url);
    
    return makeAPIRequest({
        cranium: this,
        endpoint: "/units/me/incidents",
        method: "post",
        data: new CraniumIncident(incident)
    });
};

Cranium.prototype.setAvailable = function setAvailable (value)
{
    this.available.set(value);
};

Cranium.prototype.isAvailable = function isAvailable ()
{
    return this.available.get();
};

Cranium.prototype.toJSON = function toJSON ()
{
    return _(this).omit(["available"]).value();
};

function CraniumIncident (incident)
{
    var t = _(this).extend(Objects.toJS({
        message: incident.message,
        timestamp: incident.timestamp,
        level: incident.level.toString(),
        internal: incident.internal,
        metadata: JSON.parse(cJSON.stringify(incident.metadata).getVal()),
    }, true)).value();
    
    var metadata = cJSON.stringify(incident.metadata).getVal();
    if (metadata)
    {
        t.metadata = JSON.parse(metadata);
    }
    
    if (incident.exception)
    {
        t.exception = Objects.toJS(ExceptionUtils.convertToMap(incident.exception), true);
    }
}

Cranium.FEATURES = FEATURES;

module.exports = Cranium;