/* jshint undef: true, unused: true */
/* global require, com, module, java */

var facade = require("facade");
var SystemInformationUtilities = com.cattura.agents.configuration_module.model.utilities.SystemInformationUtilities;
var TemplateHelper = com.cattura.agents.configuration_module.model.vos.capture.helpers.TemplateHelper;
var MediaTypes = com.cattura.cattura_media_commons.properties.MediaTypes;
var Processor = com.cattura.share.processor.Processor;
var Objects = require("convenience").Objects;
var _ = require("lodash");

var PluginHelper = require("lib/pluginHelper");
var PluginProxy = com.cattura.agents.configuration_module.model.proxies.PluginProxy;

var LIVE_PREVIEW_PLUGIN = "com.cattura.agents.hardware.live_device_preview_plugin.LiveDevicePreviewPlugin";

module.exports = {
    getUnitInfo: function getUnitInfo ()
    {
        var settings = facade.retrieveProxy("SettingsProxy");
        var updaterProxy = facade.retrieveProxy("UpdaterProxy");
        var versionInfo = updaterProxy.getVersionInfo();

        return Objects.toJS({
            unitName: settings.get("unit-name"),
            machineName: SystemInformationUtilities.getMachineName(),
            uuid: SystemInformationUtilities.getSystemUUID(),
            agentVersion: versionInfo.getCurrentVersion(),
            serial: settings.getPrivateSettings().get("serial")
        },true);
    },
    
    getNetworkInfo: function getNetworkInfo ()
    {
        var environment = facade.retrieveProxy("EnvironmentProxy");
        var networkWatcher = environment.networkStatusWatcher;
        return Objects.toJS({
           internet: {
               connected: networkWatcher.hasInternet(),
               ip: networkWatcher.globalIP
           },
           local: {
               ip: networkWatcher.localIP,
               type: networkWatcher.networkType,
               subnet: networkWatcher.subnet,
               gateway: networkWatcher.gateway,
               mac: networkWatcher.getMACAddress()
           }
        },true);
    },
    
    getDiskSpaceInfo: function getDiskSpaceInfo ()
    {
        var environment = facade.retrieveProxy("EnvironmentProxy");
        var diskSpaceWatcher = environment.diskSpaceWatcher;
        return Objects.toJS({
            total: diskSpaceWatcher.totalSpace,
            usable: diskSpaceWatcher.usableSpace,
            free: diskSpaceWatcher.freeSpace,
            used: diskSpaceWatcher.usedSpace
        },true);
    },
    
    getConnectedDevices: function getConnectedDevices ()
    {
        var deviceProxy = facade.retrieveProxy("DeviceProxy");
        var pluginProxy = facade.retrieveProxy(PluginProxy.NAME);
        
        var livePreviewPlugin = pluginProxy.getPluginInfo(LIVE_PREVIEW_PLUGIN).orNull();
        var videoDevicePreviewAvailable = false;
        var audioDevicePreviewAvailable = false;
        if (livePreviewPlugin)
        {
            var pluginHelper = new PluginHelper(livePreviewPlugin);
            videoDevicePreviewAvailable = pluginHelper.services.isVideoLivePreviewAvailable();
            audioDevicePreviewAvailable = pluginHelper.services.isAudioLivePreviewAvailable();
        }
        
        var deviceStructure = [];
        var audioDevices = Objects.toJS(deviceProxy.audioDevices);
        var videoDevices = Objects.toJS(deviceProxy.videoDevices);
        var key;
        var device;
        
        for (key in audioDevices)
        {
            device = {
                id: audioDevices[key].name,
                name: audioDevices[key].prettyName,
                type: "audio",
                tested: audioDevices[key].tested,
                usable: audioDevices[key].usable
            };
            
            if (audioDevicePreviewAvailable)
            {
                device.liveStream = "live/" + audioDevices[key].sanitizedID;
            }
            
            deviceStructure.push(device);
        }
        
        for (key in videoDevices)
        {
            device = {
                id: videoDevices[key].name,
                name: videoDevices[key].prettyName,
                type: "video",
                snapshot: "thumbnail/" + key,
                width: videoDevices[key].width,
                height: videoDevices[key].height,
                fps: videoDevices[key].FPS,
                tested: videoDevices[key].tested,
                usable: videoDevices[key].usable                
            };
            
            if (videoDevicePreviewAvailable)
            {
                device.liveStream =  "live/" + videoDevices[key].sanitizedID;
            }
            
            deviceStructure.push(device);
        }
        
        return Objects.toJS(deviceStructure, true);
    },
    
    getCaptureSessions: function getCaptureSessions ()
    {
        var captureStateProxy = facade.retrieveProxy("CaptureStateProxy");
        return _(Objects.toJS(captureStateProxy.captureState.allSessions, true)).map(function (session) {
            var retval = {
                state: session.state.name(),
                id: session.captureRequest.mediaPackageID,
                title: session.captureRequest.title
            };
            
            if (session.session)
            {
                _(retval).extend({
                    startTimeMillis: session.session.startTime.time,
                    elapsedMillis: session.session.elapsedTimeMillis,
                }).value();
                
                if (session.session.stopped)
                {
                    _(retval).extend({
                        stopTimeMillis: session.session.stopTime.time
                    }).value();
                }
            }
            
            return Objects.toJS(retval, true);
        }).value();
    },
    
    getTemplates: function getTemplates ()
    {
        var templates = facade.retrieveProxy("CaptureTemplateProxy");
        var mediaProcessing = facade.retrieveProxy("MediaProcessingProxy").getCaptureModule();
        return _(Objects.toJS(templates.templates, true)).map(function (template) {
            var helper = new TemplateHelper(template);
            return Objects.toJS({
                id: template.ID,
                title: template.title,
                audioSources: _(Objects.toJS(helper.getSources(), true)).filter(function (source) {
                    return source.type === MediaTypes.Device.AUDIO;
                }).map(function (source) {
                    return {
                        id: source.sourceId
                    };
                }).value(),
                videoSources: _(Objects.toJS(helper.getSources(), true)).filter(function (source) {
                    return source.type === MediaTypes.Device.VIDEO;
                }).map(function (source) {
                    return {
                        id: source.sourceId
                    };
                }).value(),
                destinations: _(Objects.toJS(template.getGraph().selectNodesByType(Processor.TYPE.INTEGRATION))).map(function (node) {
                    var plugin = mediaProcessing.getPlugin(node.pluginID());
                    return {
                        id: node.pluginID(),
                        title: plugin ? plugin.getTitle() : node.getPluginID()
                    };
                }).value()
            }, true);
        }).value();
    },
    
    getSettings: function getSettings ()
    {
        var settings = facade.retrieveProxy("SettingsProxy");
        
        return Objects.toJS(settings.getUserSettings().toMap(), true);
    },
    
    observeCaptureState: function observeCaptureState (fn)
    {
        var captureStateProxy = facade.retrieveProxy("CaptureStateProxy");
        captureStateProxy.getCaptureState().addObserver(new java.util.Observer({
            update: function update (captureState, sessionState)
            {
                fn(captureState, sessionState);
            }
        }));
    }
};
