/* jshint undef: true, unused: true */
/* global plugin, require, com, org, java */

var _ = require("lodash"),
    on = require("on")(plugin),
    Objects = require("convenience").Objects,
    schedule = require("lib/schedule"),
    status = require("lib/status"),
    Cranium = require("lib/cranium"),
    getIPAddress = require("lib/util").getIPAddress,
    EventType = com.cattura.agents.configuration_module.model.vos.events.EventType,
    templates = {
        authorizev1: plugin.resource("templates", "authorize.tmpl"),
        authorizev2: plugin.resource("templates", "authorizev2.tmpl")
    },
    STORE = plugin.info().dataStore,
    bearerPrefix = "bearer ",
    facade = require("facade"),
    IncidentReport = com.cattura.agents.configuration_module.model.vos.IncidentReport,
    
    /*
     * Monitoring and reporting on status of blocked tasks
     */
    FutureStatusWatcher = com.cattura.agents.configuration_module.model.vos.watchers.FutureStatusWatcher,
    Time = com.cattura.share.utils.time.Time,
    Duration = java.time.Duration,
    ChronoUnit = java.time.temporal.ChronoUnit,
    
    /*
     * Logging
     */
    logger = plugin.info().logger,
    console = require("console")(logger),
    Level = java.util.logging.Level,
    
    /*
     * Executor used to handle events and push updates to Cranium
     */
    ThreadFactoryBuilder = com.google.common.util.concurrent.ThreadFactoryBuilder,
    eventExecutor = java.util.concurrent.Executors.newSingleThreadExecutor(new ThreadFactoryBuilder()
        .setNameFormat("Cranium Plugin event pusher thread %d")
        .setDaemon(true)
    .build()),
    
    /*
     * Executor used to spawn other threads as needed
     */
    executor = java.util.concurrent.Executors.newCachedThreadPool(new ThreadFactoryBuilder()
        .setNameFormat("Cranium Plugin executor thread %d")
        .setDaemon(true)
    .build()),
    
    /*
     * Incidents that we send out. This is to avoid handling our own incidents
     * in our IncidentEvent handler.
     */
    ourIncidents = [],
    
    /*
     * Instances of Cranium that we are connected to
     */
    craniums;

/*
 * Load Craniums from data store
 */
craniums = _(Objects.toJS(STORE.select("craniums[*]"))).map(function (cranium) {
    return new Cranium(JSON.parse(cranium));
}).value();

logger.log(Level.FINE, "Loaded " + craniums.length + " craniums");

/*
 * By default the CIP will broadcast anything that's not an internal event.
 */
function eventIsNotifiable (event)
{
    if (event.type.category == EventType.Category.INTERNAL ||
        event.type == EventType.EXCEPTION)
    {
        return false;
    }
    
    return true;
}

/*
 * Register an authorized cranium instance. We will be able to send
 * messages to it, and recieve messages from it.
 */
function addCranium (cranium)
{
    cranium = new Cranium(Objects.toJS(cranium, true));
    cranium.id = craniums.length;
    STORE.set("craniums[" + cranium.id + "]", JSON.stringify(cranium));
    logger.log(Level.INFO, "Added cranium (id: " + cranium.id + ", url: " + cranium.url + ")");
    craniums.push(cranium);
    return cranium;
}

/*
 * Remove a cranium instance from our data store. This will
 * cause us to no longer send that cranium any status updates.
 * This is invoked when the cranium user selects "Remove Unit" from
 * the cranium UI.
 */
function removeCranium (cranium)
{
    STORE.remove("craniums[" + cranium.id + "]");
    logger.log(Level.INFO, "Removed cranium (id: " + cranium.id + ", url: " + cranium.url + ")");
    craniums = _(craniums).without(cranium).value();
}

/*
 * Get "available" cranium instances, i.e.
 * those that are online and that we are able
 * to send messages to.
 * 
 * A cranium is marked as unavailable if we fail to
 * connect to it, either during initialization or
 * during the lifetime of the plugin. If this happens, 
 * the cranium must send a message notifying us that 
 * it is available again in order for us to begin 
 * sending messages again.
 */
function getAvailableCraniums ()
{
    return _(craniums).filter(function (cranium) {
        return cranium.isAvailable();
    });
}

function getSettings ()
{
    return {};
}

function setSettings ()
{
    return getSettings();
}

function getCraniumForRequest (req)
{
    var authorization,
        token,
        cranium;
    
    if (!req.hasHeader("Authorization")) return null;
    authorization = req.header("Authorization");
    if (!authorization.startsWith(bearerPrefix)) return null;
    
    token = Objects.toJS(authorization.substring(bearerPrefix.length), true);
    cranium = _(craniums).find(function (cranium) {
        return token === cranium.inboundAccessToken;
    });
    
    if (!cranium)
    {
        logger.log(Level.FINE, "Unable to find cranium instance (auth header: " + authorization + ")");
    }
    
    return cranium;
}

function rejectRequest (req, res)
{
    console.log("Rejected request from " + getIPAddress(req) + " due to invalid authorization");
    console.log(" - " + req.header("Authorization"));
    res.status(401).header("Content-Type", "application/json").content(JSON.stringify({err: "Unauthorized"})).end();
}

/*
 * Decorator for internal (Cranium -> CCCP) API functions.
 */
function apiHandler (func)
{
    return new org.webbitserver.HttpHandler({
        handleHttpRequest: function (req, res) {
            var cranium = getCraniumForRequest(req);
            if (!cranium)
            {
                rejectRequest(req, res);
                return;
            }
            
            if (!cranium.isAvailable())
            {
                logger.log(Level.INFO, "Cranium instance at " + cranium.url + " is available");
            }
            
            cranium.setAvailable(true);
            func(req, res, cranium);
        }
    });
}

plugin.service({
    name: "settings.get",
    description: "Get the settings of the Cranium Integration Plugin",
    parameters: [],
    handler: getSettings
});

plugin.service({
    name: "settings.set",
    description: "Set the settings of the Cranium Integration Plugin Plugin",
    parameters: [{name: "settings", description: "Settings to set", type: "java.util.Map"}],
    handler: setSettings
});

function handleError (err, message)
{
    var exception = err ? (err.javaException || new java.lang.Exception(err.message)) : null;
    message = message || "Failed to connect to Command Center";
    var incident = new IncidentReport(message, exception);
    incident.setInternal(true);
    ourIncidents.push(incident);
    facade.sendNotification(facade.HANDLE_EXCEPTION, incident);
}

function runOnExecutor (title, fn)
{
    var future = eventExecutor.submit(new java.lang.Runnable({
        run: function () {
            try
            {
                fn();
            }
            catch (err)
            {
                handleError(err, "Uncaught error in Command Center plugin from task \"" + title + "\"");
            }
        }
    }));
    
    executor.submit(new FutureStatusWatcher(future, Duration.of(5, ChronoUnit.MINUTES), 5, new java.lang.Runnable({
        run: function run ()
        {
            handleError(null, new IncidentReport("Command Center update thread task \"" + title + "\" timed out"));
            future.cancel(true);
        }
    })));
    
    return future;
}

/*
 * This thread sends Cranium an update at the start of a new day, to keep
 * schedule updates current.
 */
executor.execute((function () {
    var dateFormatter = new java.text.SimpleDateFormat("y-M-d"),
        currentDate = dateFormatter.format(new java.util.Date()),
        intervalMillis = 1800000; // 30 minutes
    
    return new java.lang.Runnable({
        run: function () {
            console.log("Daily update thread starting");
            
            while (true)
            {
                var nowString = dateFormatter.format(new java.util.Date());
                if (nowString.equalsIgnoreCase(currentDate))
                {
                    Time.sleep(intervalMillis);
                    continue;
                }
                
                currentDate = nowString;
                console.log("Pushing out daily update");
                
                /* jshint loopfunc:true */
                _(craniums).each(function (cranium) {
                    cranium.checkAvailability().then(function () {
                        if (cranium.isAvailable())
                        {
                            cranium.pushInitialUpdate().catch(handleError);
                        }
                    });
                });
                /* jshint loopfunc:false */
            }
        }
    }); 
})());

/*
 * Push capture session status update. Additionally, if the given mediaPackageIds
 * correspond to scheduled events, those are pushed too. If no mediaPackageIds
 * are given, it will fetch the list of currently running capture sessions and
 * derive mediaPackageIds from that.
 */
function pushCaptureStatusUpdate (mediaPackageIds)
{
    var eventsToUpdate,
        availableCraniums = getAvailableCraniums();
    
    if (!mediaPackageIds)
    {
        mediaPackageIds = _(status.getCaptureSessions()).map(function (session) {
            return session.id;
        }).value();
    }
    
    eventsToUpdate =  _(schedule.getSchedule().events).filter(function (scheduleEvent) {
        return _(mediaPackageIds).includes(scheduleEvent.mediaPackageID);
    }).value();
    
    logger.log(Level.FINE, "Pushing capture state update (" + availableCraniums.length + " craniums to be notified)");
    runOnExecutor("Push capture state update", function () {
        _(availableCraniums).each(function (cranium) {
            cranium.pushCaptureSessionsUpdate().catch(handleError);

            _(eventsToUpdate).each(function (event) {
                if (event)
                {
                    cranium.pushScheduleEventUpdate(event).catch(handleError);
                }
            });
        });
    });    
}

/*
 * Attach observer to CaptureState object for now, since the CaptureStatusEvent
 * fires too often to be of any practical use (it fires whenever the state of an individual node
 * changes, which happens way too often), whereas the CaptureState observer only fires 
 * when the overall state is updated.
 */
status.observeCaptureState(function (captureState) {
    pushCaptureStatusUpdate(_(Objects.toJS(captureState.getAllSessions())).map(function (sessionState) {
        return "" + sessionState.getCaptureRequest().getMediaPackageID();
    }).value());
});

on({
    /*
     * Push initial update to Cranium when all Configuration Module
     * plugins are loaded.
     */
    WebServerStartedEvent: function () {
        logger.log(Level.FINE, "Pushing initial status updates (" + craniums.length + " craniums to be notified)");
        runOnExecutor("Push initial status update", function () {
            _(craniums).each(function (cranium) {
                cranium.checkAvailability().then(function () {
                    if (cranium.isAvailable())
                    {
                        cranium.pushInitialUpdate().catch(handleError);
                    }
                });
            });
        });
    },
    
    /*
     * Push general events to Cranium
     */
    Event: function (event) {
        if (eventIsNotifiable(event))
        {
            var availableCraniums = getAvailableCraniums();
            logger.log(Level.FINE, "Pushing event update (event type " + event.getClass().getSimpleName() + "; " + availableCraniums.length + " craniums to be notified)");        
            runOnExecutor("Push event", function () {
                _(availableCraniums).each(function (cranium) {
                    cranium.pushEvent(event).catch(handleError);
                });
            });
        }
    },
    
    /*
     * Push media package related events. This is start/stop/failed etc.
     */
    MediaPackageEvent: function (event) {
        // Don't handle CaptureStatusEvent here
        if (event.getClass().getSimpleName() == "CaptureStatusEvent")
        {
            return;
        }
        
        if (event.getMediaPackageID())
        {
            pushCaptureStatusUpdate(["" + event.getMediaPackageID()]);
        }
    },
    
    /*
     * Push capture session update
     * Do not use this event since it fires too often right now.
     */
    /*CaptureStatusEvent: function (event) {
        
    },*/
    
    /*
     * Push device detected events to Cranium
     */
    DevicesDetectedEvent: function () {
        var availableCraniums = getAvailableCraniums();
        logger.log(Level.FINE, "Pushing devices detected update (" + availableCraniums.length + " craniums to be notified)");
        runOnExecutor("Push devices update", function () {
            _(availableCraniums).each(function (cranium) {
                cranium.pushDevicesUpdate().catch(handleError);
            });
        });
    },
    
    /*
     * Push template created/destroy events to Cranium
     */
    TemplateCreatedEvent: function () {
        var availableCraniums = getAvailableCraniums();
        logger.log(Level.FINE, "Pushing template created update (" + availableCraniums.length + " craniums to be notified)");
        runOnExecutor("Push template update", function () {
            _(availableCraniums).each(function (cranium) {
                cranium.pushTemplatesUpdate().catch(handleError);
            });
        });
    },
    
    /*
     * Push device detected events to Cranium
     */
    TemplateDeletedEvent: function () {
        var availableCraniums = getAvailableCraniums();
        logger.log(Level.FINE, "Pushing template deleted update (" + availableCraniums.length + " craniums to be notified)");
        runOnExecutor("Push template update", function () {
            _(availableCraniums).each(function (cranium) {
                cranium.pushTemplatesUpdate().catch(handleError);
            });
        });
    },
    
    /*
     * Push schedule events to Cranium
     */
    ScheduledEventsEvent: function () {
        var availableCraniums = getAvailableCraniums();
        logger.log(Level.FINE, "Pushing schedule update (" + availableCraniums.length + " craniums to be notified)");
        runOnExecutor("Push schedule update", function () {
            _(availableCraniums).each(function (cranium) {
                cranium.pushScheduleUpdate().catch(handleError);
            });
        });
    },
    
    /*
     * Push schedule update and event source update
     */
    SchedulerEventSourceConfiguredEvent: function () {
        var availableCraniums = getAvailableCraniums();
        logger.log(Level.FINE, "Pushing scheduler configuration update (" + availableCraniums.length + " craniums to be notified)");
        runOnExecutor("Push scheduler configuration update", function () {
            _(availableCraniums).each(function (cranium) {
                cranium.pushSchedulerConfigUpdate().then(function () { 
                    return cranium.pushScheduleUpdate();
                }).catch(handleError);
            });
        });
    },
    
    SchedulerEventSourceSyncEvent: function () {
        var availableCraniums = getAvailableCraniums();
        logger.log(Level.FINE, "Pushing schedule update (" + availableCraniums.length + " craniums to be notified)");
        runOnExecutor("Push schedule update", function () {
            _(availableCraniums).each(function (cranium) {
                cranium.pushScheduleUpdate().catch(handleError);
            });
        });
    },
    
    /*
     * When user settings (e.g. label) are updated, push a status update
     * to Cranium to keep the info on that side synced.
     */
    SettingsUpdatedEvent: function () {
        var availableCraniums = getAvailableCraniums();
        logger.log(Level.FINE, "Pushing unit info update (" + availableCraniums.length + " craniums to be notified)");
        runOnExecutor("Push unit info update", function () {
            _(availableCraniums).each(function (cranium) {
                cranium.pushStatusUpdate().then(function () { 
                    return cranium.pushSettingsUpdate();
                }).catch(handleError);
            });
        });
    },
    
    /*
     * When an incident occurs, notify Cranium.
     */
    IncidentEvent: function (event) {
        var availableCraniums = getAvailableCraniums(),
            incident = event.getIncident();
        
        /*
         * If this incident is one we generated, we don't want to handle it here.
         * This is because our incidents are usually due to a failure to connect,
         * so trying to push them can result in an infinite loop.
         */
        if (_(ourIncidents).includes(incident))
        {
            logger.log(Level.FINE, "Not pushing incident \"" + event.incident.message + "\" because it's one of our own");
            return;
        }
        
        logger.log(Level.FINE, "Pushing incident \"" + event.incident.message + "\" (" + availableCraniums.length + " craniums to be notified)");
        runOnExecutor("Push incident", function () {
            _(availableCraniums).each(function (cranium) {
                cranium.pushIncident(incident).catch(handleError);
            });
        });
    },
    
    /*
     * Builds the interface for Cranium to communicate with CCCP.
     */
    WebServerEvent: function (event) {
        var HttpHandler = org.webbitserver.HttpHandler,
            server = event.webServer,
            rest = new org.webbitserver.rest.Rest(server);
        
        /*
         * Used by Cranium to confirm that the unit is compatible with
         * Cranium and can be connected to.
         * 
         * The "version" field corresponds to the version of the "internalApi" 
         * feature and is deprecated.
         */
        rest.GET("/cranium", new HttpHandler({
            handleHttpRequest: function (req, res) {
                res.status(200).header("Content-Type", "application/json").content(JSON.stringify(Objects.toJS({
                    result: "ok",
                    version: Cranium.FEATURES.internalApi,
                    ip: getIPAddress(req),
                    features: Cranium.FEATURES
                }, true))).end();
            }
        }));
        
        /*
         * Set up new auth flow.
         */
        require("flow/v2")({
            server: server,
            authorizeTemplate: templates.authorizev2,
            addCranium: addCranium,
            logger: logger
        });
        
        /*
         * Set up legacy auth flow.
         */
        require("flow/legacy")({
            server: server,
            authorizeTemplate: templates.authorizev1,
            craniums: craniums,
            addCranium: addCranium
        });
        
        /*
         * Ping endpoint for Cranium, so Cranium can keep check on the unit's
         * connectivity.
         */
        server.add("/cranium/ping", apiHandler(function (req, res) {
            res.status(200).header("Content-Type", "application/json").content(JSON.stringify(Objects.toJS({result: "ok"}, true))).end();            
        }));
        
        /*
         * Endpoint for Cranium to request a full update from the unit.
         */
        server.add("/cranium/requestUpdate", apiHandler(function (req, res, cranium) {
            cranium.pushInitialUpdate().then(function () {
                res.status(200).header("Content-Type", "application/json").content(JSON.stringify(Objects.toJS({result: "ok"}, true))).end();
            });
        }));
        
        /*
         * Endpoint for Cranium to push a schedule update to the unit.
         * Returns the updated schedule.
         */
        rest.PUT("/cranium/schedule", apiHandler(function (req, res) {
            var scheduleData = JSON.parse(req.body());
            
            res.status(200).header("Content-Type", "application/json").content(JSON.stringify(Objects.toJS({
                result: "ok",
                schedule: schedule.updateSchedule(scheduleData)
            }, true))).end();
        }));
        
        /*
         * Either creates a new event or performs a batch schedule update.
         */
        rest.POST("/cranium/schedule", apiHandler(function (req, res) {
            var scheduleData = JSON.parse(req.body());

            if (scheduleData.range && scheduleData.events)
            {
                res.status(200).header("Content-Type", "application/json").content(JSON.stringify(Objects.toJS({
                    result: "ok",
                    schedule: schedule.updateSchedule(scheduleData)
                }, true))).end();
            }
            else
            {
                res.status(200).header("Content-Type", "application/json").content(JSON.stringify(Objects.toJS({
                    result: "ok",
                    event: schedule.createEvent(scheduleData)
                }, true))).end();
            }
        }));
        
        /*
         * Endpoint for Cranium to update a single scheduled event.
         * Returns the updated event.
         */
        rest.PUT("/cranium/schedule/{eventId}", apiHandler(function (req, res) {
            var eventData = JSON.parse(req.body());
            eventData.id = rest.param(req, "eventId");
            res.status(200).header("Content-Type", "application/json").content(JSON.stringify(Objects.toJS({
                result: "ok",
                event: schedule.updateEvent(eventData)
            }, true))).end();
        }));
        
        /*
         * Endpoint for Cranium to delete a single scheduled event.
         * Returns the updated event.
         */
        rest.DELETE("/cranium/schedule/{eventId}", apiHandler(function (req, res) {
            res.status(200).header("Content-Type", "application/json").content(JSON.stringify(Objects.toJS({
                result: "ok",
                event: schedule.deleteEvent({
                    id: rest.param(req, "eventId")
                })
            }, true))).end();
        }));
        
        /*
         * Deletes the record of the requesting Cranium instance.
         * Used when Cranium wants to unlink the unit from itself.
         */
        rest.POST("/cranium/unlink", apiHandler(function (req, res, cranium) {
            removeCranium(cranium);
            res.status(200).header("Content-Type", "application/json").content(JSON.stringify(Objects.toJS({
                result: "ok"
            }, true))).end();
        }));
    },
    
    /*
     * Register our authenticator to the API builder. This allows Cranium to
     * make authenticated API requests using its token.
     */
    BuildAPIEvent: function (event) {
        var Authenticator = com.cattura.agents.hardware.api.vos.APIBuilder.Authenticator;
        var Optional = java.util.Optional;
        
        event.builder.addAuthenticator(new Authenticator({
            authenticate: function authenticate (req) {
                var cranium = getCraniumForRequest(req);
                if (!cranium)
                {
                    return Optional.empty();
                }
                
                return Optional.of(cranium.asUser());
            }
        }));
    }
});
